package types

import (
	"fmt"

	"github.com/hashicorp/terraform/configs/configschema"
	"github.com/mcuadros/ascode/terraform"
	"go.starlark.net/starlark"
	"go.starlark.net/syntax"
)

type ResourceCollection struct {
	typ      string
	kind     Kind
	block    *configschema.Block
	provider *Provider
	parent   *Resource
	*starlark.List
}

func NewResourceCollection(
	typ string, k Kind, block *configschema.Block, provider *Provider, parent *Resource,
) *ResourceCollection {
	return &ResourceCollection{
		typ:      typ,
		kind:     k,
		block:    block,
		provider: provider,
		parent:   parent,
		List:     starlark.NewList(nil),
	}
}

// LoadList loads a list of dicts on the collection. It clears the collection.
func (c *ResourceCollection) LoadList(l *starlark.List) error {
	if err := c.List.Clear(); err != nil {
		return err
	}

	for i := 0; i < l.Len(); i++ {
		dict, ok := l.Index(i).(*starlark.Dict)
		if !ok {
			return fmt.Errorf("%d: expected dict, got %s", i, l.Index(i).Type())
		}

		c.MakeResource("", dict)
	}

	return nil
}

// String honors the starlark.Value interface.
func (c *ResourceCollection) String() string {
	return fmt.Sprintf("%s", c.typ)
}

// Type honors the starlark.Value interface.
func (c *ResourceCollection) Type() string {
	return fmt.Sprintf("ResourceCollection<%s.%s>", c.kind, c.typ)
}

// Truth honors the starlark.Value interface.
func (c *ResourceCollection) Truth() starlark.Bool {
	return true // even when empty
}

// Freeze honors the starlark.Value interface.
func (c *ResourceCollection) Freeze() {}

// Hash honors the starlark.Value interface.
func (c *ResourceCollection) Hash() (uint32, error) {
	return 0, fmt.Errorf("unhashable type: ResourceCollection")
}

// Name honors the starlark.Callable interface.
func (c *ResourceCollection) Name() string {
	return c.typ
}

// CallInternal honors the starlark.Callable interface.
func (c *ResourceCollection) CallInternal(thread *starlark.Thread, args starlark.Tuple, kwargs []starlark.Tuple) (starlark.Value, error) {
	name, dict, err := c.unpackArgs(args, kwargs)
	if err != nil {
		return nil, err
	}

	return c.MakeResource(name, dict)
}

// MakeResource it makes a new resource and loads the dict on it.
func (c *ResourceCollection) MakeResource(name string, dict *starlark.Dict) (*Resource, error) {
	if (c.kind == ResourceKind || c.kind == DataSourceKind) && name == "" {
		name = NameGenerator()
	}

	resource := MakeResource(name, c.typ, c.kind, c.block, c.provider, c.parent)
	if dict != nil && dict.Len() != 0 {
		if err := resource.LoadDict(dict); err != nil {
			return nil, err
		}
	}

	if err := c.List.Append(resource); err != nil {
		return nil, err
	}

	return resource, nil
}

func (c *ResourceCollection) unpackArgsWithKwargs(args starlark.Tuple, kwargs []starlark.Tuple) (string, *starlark.Dict, error) {
	dict := starlark.NewDict(len(kwargs))
	var name starlark.String

	for _, kwarg := range kwargs {
		dict.SetKey(kwarg.Index(0), kwarg.Index(1))
	}

	if len(args) == 1 {
		var ok bool
		name, ok = args.Index(0).(starlark.String)
		if !ok {
			return "", nil, fmt.Errorf("resource: expected string, got %s", args.Index(0).Type())
		}
	}

	if len(args) > 1 {
		return "", nil, fmt.Errorf("resource: unexpected positional args mixed with kwargs")
	}

	return string(name), dict, nil
}

func (c *ResourceCollection) unpackArgs(args starlark.Tuple, kwargs []starlark.Tuple) (string, *starlark.Dict, error) {
	var dict *starlark.Dict
	var name starlark.String

	if len(args) == 0 && len(kwargs) == 0 {
		return "", nil, nil
	}

	if len(kwargs) != 0 {
		return c.unpackArgsWithKwargs(args, kwargs)
	}

	switch len(args) {
	case 0:
	case 1:
		switch v := args.Index(0).(type) {
		case starlark.String:
			return string(v), nil, nil
		case *starlark.Dict:
			return "", v, nil
		default:
			return "", nil, fmt.Errorf("resource: expected string or dict, got %s", args.Index(0).Type())
		}
	case 2:
		var ok bool
		name, ok = args.Index(0).(starlark.String)
		if !ok {
			return "", nil, fmt.Errorf("resource: expected string, got %s", args.Index(0).Type())
		}

		dict, ok = args.Index(1).(*starlark.Dict)
		if !ok {
			return "", nil, fmt.Errorf("resource: expected dict, got %s", args.Index(1).Type())
		}
	default:
		if c.kind != NestedKind {
			return "", nil, fmt.Errorf("resource: unexpected positional arguments count")
		}
	}

	return string(name), dict, nil
}

func (c *ResourceCollection) toDict() *starlark.List {
	values := make([]starlark.Value, c.Len())
	for i := 0; i < c.Len(); i++ {
		values[i] = c.Index(i).(*Resource).toDict()
	}

	return starlark.NewList(values)
}

// Attr honors the starlark.HasAttrs interface.
func (c *ResourceCollection) Attr(name string) (starlark.Value, error) {
	switch name {
	case "search":
		return starlark.NewBuiltin("search", c.search), nil
	case "__dict__":
		return c.toDict(), nil
	}

	return c.List.Attr(name)
}

// AttrNames honors the starlark.HasAttrs interface.
func (c *ResourceCollection) AttrNames() []string {
	return append(c.List.AttrNames(), "search", "__dict__")
}

func (c *ResourceCollection) search(_ *starlark.Thread, _ *starlark.Builtin, args starlark.Tuple, _ []starlark.Tuple) (starlark.Value, error) {
	var key string
	var value starlark.Value

	switch len(args) {
	case 1:
		key = "id"
		value = args.Index(0)
	case 2:
		string, ok := args.Index(0).(starlark.String)
		if !ok {
			return nil, fmt.Errorf("resource: expected string, got %s", args.Index(0).Type())
		}

		key = string.GoString()
		value = args.Index(1)
	default:
		return nil, fmt.Errorf("search: unexpected positional arguments count")
	}

	list := starlark.NewList(nil)
	for i := 0; i < c.Len(); i++ {
		r := c.Index(i).(*Resource)
		v, ok := getValue(r, key).(starlark.Comparable)
		if !ok || v.Type() != value.Type() {
			continue
		}

		match, err := v.CompareSameType(syntax.EQL, value, 2)
		if err != nil {
			return starlark.None, err
		}

		if match {
			list.Append(r)
		}
	}

	return list, nil
}

func getValue(r *Resource, key string) starlark.Value {
	if key == "id" {
		return starlark.String(r.name)
	}

	if !r.values.Has(key) {
		return starlark.None
	}

	return r.values.Get(key).Starlark()
}

type ProviderCollection struct {
	pm *terraform.PluginManager
	*AttrDict
}

func NewProviderCollection(pm *terraform.PluginManager) *ProviderCollection {
	return &ProviderCollection{
		pm:       pm,
		AttrDict: NewAttrDict(),
	}
}

// String honors the starlark.Value interface.
func (c *ProviderCollection) String() string {
	return "foo"
}

// Type honors the starlark.Value interface.
func (c *ProviderCollection) Type() string {
	return "ProviderCollection"
}

// Truth honors the starlark.Value interface.
func (c *ProviderCollection) Truth() starlark.Bool {
	return true // even when empty
}

// Freeze honors the starlark.Value interface.
func (c *ProviderCollection) Freeze() {}

// Hash honors the starlark.Value interface.
func (c *ProviderCollection) Hash() (uint32, error) {
	return 0, fmt.Errorf("unhashable type: ProviderCollection")
}

// Name honors the starlark.Callable interface.
func (c *ProviderCollection) Name() string {
	return "foo"
}

// CallInternal honors the starlark.Callable interface.
func (c *ProviderCollection) CallInternal(thread *starlark.Thread, args starlark.Tuple, kwargs []starlark.Tuple) (starlark.Value, error) {
	name, version, alias, err := c.unpackArgs(args)
	if err != nil {
		return nil, err
	}

	return c.MakeProvider(name, version, alias, kwargs)
}

func (c *ProviderCollection) MakeProvider(name, version, alias string, kwargs []starlark.Tuple) (*Provider, error) {
	n := starlark.String(name)
	a := starlark.String(alias)

	if _, ok, _ := c.Get(n); !ok {
		c.SetKey(n, NewAttrDict())
	}
	providers, _, _ := c.Get(n)
	if _, ok, _ := providers.(*AttrDict).Get(a); ok {
		return nil, fmt.Errorf("already exists a provider %q with the alias %q", name, alias)
	}

	p, err := MakeProvider(c.pm, name, version, alias)
	if err != nil {
		return nil, err
	}

	if err := providers.(*AttrDict).SetKey(starlark.String(p.alias), p); err != nil {
		return nil, err
	}

	return p, p.loadKeywordArgs(kwargs)
}

func (c *ProviderCollection) unpackArgs(args starlark.Tuple) (string, string, string, error) {
	var name, version, alias starlark.String
	switch len(args) {
	case 3:
		var ok bool
		alias, ok = args.Index(2).(starlark.String)
		if !ok {
			return "", "", "", fmt.Errorf("expected string, got %s", args.Index(2).Type())
		}
		fallthrough
	case 2:
		var ok bool
		version, ok = args.Index(1).(starlark.String)
		if !ok {
			return "", "", "", fmt.Errorf("expected string, got %s", args.Index(1).Type())
		}
		fallthrough
	case 1:
		var ok bool
		name, ok = args.Index(0).(starlark.String)
		if !ok {
			return "", "", "", fmt.Errorf("expected string, got %s", args.Index(0).Type())
		}
	default:
		return "", "", "", fmt.Errorf("unexpected positional arguments count")
	}

	return string(name), string(version), string(alias), nil
}
